#ifndef _ARCFS_H
#define _ARCFS_H

#define ARC_ROOT_INODE 1

struct arcfs_entry {
  struct arcfs_entry *next, *sub;
  struct super_block *sb;

  unsigned long		ino;
  kdev_t		dev;
  umode_t		mode;
  nlink_t		nlink;
  uid_t			uid;
  gid_t			gid;
  kdev_t		rdev;
  time_t		atime;
  time_t		mtime;
  time_t		ctime;
  char			*symlink;

  /* Follows must be filled in when install_arcfs() */
  int			type;		/* Entry type. 0 for no conversion */
  off_t			size;		/* Original archive file size */
  off_t			compsize;	/* Compressed archive file size */
  off_t			offset;		/* Offset start point */

  void *private_data;	/* Arc module private */
  char	name[0];	/* filename */
};

struct arcfs_super_block
{
  struct arcfs_entry *root;
  struct arcfs_module *module;
  char *arcfile;	/* Real archive file on the disk */
  void *mem;		/* To use allocate arcfs_entry */
};


/* Compression/Encoding type (will be obsolete) */
enum
{
    ARCFS_STORED,		/* No compression */
    ARCFS_PATHNAME,		/* Pathname (Contents exists there) */
    ARCFS_COMPRESSED,		/* Compressed */
    ARCFS_PACKED,		/* Packed */
    ARCFS_DEFLATED,		/* Deflate */
    ARCFS_SHRUNKED,		/* Shrunked */
    ARCFS_REDUCED1,		/* Reduced with compression factor 1 */
    ARCFS_REDUCED2,		/* Reduced with compression factor 2 */
    ARCFS_REDUCED3,		/* Reduced with compression factor 3 */
    ARCFS_REDUCED4,		/* Reduced with compression factor 4 */
    ARCFS_IMPLODED,		/* Implode base-tag */
    ARCFS_IMPLODED_LIT8,	/* 8K sliding window (coded) */
    ARCFS_IMPLODED_LIT4,	/* 4K sliding window (coded) */
    ARCFS_IMPLODED_NOLIT8,	/* 8K sliding window (uncoded) */
    ARCFS_IMPLODED_NOLIT4,	/* 4K sliding window (uncoded) */
    ARCFS_LZHED,		/* LZH base-tag */
    ARCFS_LZHED_LH0,		/* -lh0- (ARCFS_STORED) */
    ARCFS_LZHED_LH1,		/* -lh1- */
    ARCFS_LZHED_LH2,		/* -lh2- */
    ARCFS_LZHED_LH3,		/* -lh3- */
    ARCFS_LZHED_LH4,		/* -lh4- */
    ARCFS_LZHED_LH5,		/* -lh5- */
    ARCFS_LZHED_LZS,		/* -lzs- */
    ARCFS_LZHED_LZ5,		/* -lz5- */
    ARCFS_LZHED_LZ4,		/* -lz4- (ARCFS_STORED) */
    ARCFS_LZHED_LHD,		/* -lhd- (Directory, No compression data) */
    ARCFS_LZHED_LH6,		/* -lh6- */
    ARCFS_LZHED_LH7,		/* -lh7- */

    /* Encode for MIME */
    ARCFS_UU,		/* uu encoded */
    ARCFS_B64,		/* base64 encoded */
    ARCFS_QS,		/* quoted string encoded */
    ARCFS_HQX,		/* HQX encoded */
    ARCFS_EXT		/* For extensional module */
};

struct arc_file
{
  struct arcfs_entry *entry;

  int fd;
  off_t offset;		/* offset seek pointer for `fd' */
  off_t pos;		/* current file position (0..size) */
  off_t size;		/* maxmum file position */
  int type;		/* To use decompress/decode. 0 for no conversion. */
  void *decoder;	/* converter */
  void *private_data;	/* Arc module private */
};

struct arcfs_module
{
  int (* name_check)(const char *path);
  int (* install_arcfs)(struct super_block *sb);
  void (* uninstall_arcfs)(struct super_block *sb);

  /* Archive file operations */
  struct arc_file *(* arc_open)(struct arcfs_entry *entry);
  loff_t  (* arc_seek)(struct arc_file *arc_filp, loff_t pos);
  ssize_t (* arc_read)(struct arc_file *arc_filp, char *buff, size_t len);
  int     (* arc_close)(struct arc_file *arc_filp);

  struct arcfs_module *next;			/* chain next modules */
};
/* Follows for no conversion, and read data from fd. */
extern struct arc_file *default_arc_open(struct arcfs_entry *entry);
extern loff_t default_arc_seek(struct arc_file *arc_filp, loff_t pos);
extern ssize_t default_arc_read(struct arc_file *arc_filp,
				char *buff, size_t len);
extern int default_arc_close(struct arc_file *arc_filp);

/* Call read_arc_stream() to adjust seek pointer */
extern ssize_t arc_skip_read(struct arc_file *arc_filp,
			     int (* arc_read)(struct arc_file *arc_filp,
					      char *buff, size_t len),
			     size_t len);

extern void *arcfs_alloc_memory(struct arcfs_super_block *asb, size_t len);

/* i-node <---> arcfs entry */
#define INO2ARCENT(ino)		((struct arcfs_entry *)(ino<<1))
#define ARCENT2INO(arcent)	(((unsigned long)arcent)>>1)

/* Add new archive entry */
extern struct arcfs_entry *arcfs_add_entry(struct super_block *sb,
					   char *path);
extern struct arcfs_entry *arcfs_find_entry(struct super_block *sb,
					    char *path);

extern int arcfs_register(struct arcfs_module *module);
extern int arcfs_unregister(struct arcfs_module *module);

extern struct arcfs_module *arcfs_find_module(const char *path);
extern struct arcfs_entry *arcfs_alloc_entry(struct super_block *sb,
					     char *filename, int len);
time_t arc_dos2unix_time(unsigned short time, unsigned short date);
extern char *arcfs_strip_topdir(char *fname, int *len_ret);
extern struct arcfs_module *arcfs_module_list;

#define ENTRY_ARCFS(entry) ((struct arcfs_super_block *)(entry->sb->u.generic_sbp))

/*#define DEBUG 1*/

#ifdef DEBUG
#define PRINTK(x) printk x
#else
#define PRINTK(x)
#endif


#endif /* _ARCFS_H */
