#include <linux/module.h>
#include <linux/fs.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/slab.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static int errno;

#include "arcfs.h"
#include "mblock.h"
#include "unlzh.h"

#define EOF -1

#define LZHEADER_STRAGE 4096
#define FILENAME_LENGTH 1024

#define I_HEADER_SIZE			0
#define I_HEADER_CHECKSUM		1
#define I_METHOD			2
#define I_PACKED_SIZE			7
#define I_ORIGINAL_SIZE			11
#define I_LAST_MODIFIED_STAMP		15
#define I_ATTRIBUTE			19
#define I_HEADER_LEVEL			20
#define I_NAME_LENGTH			21
#define I_NAME				22

#define EXTEND_GENERIC		0
#define EXTEND_UNIX		'U'
#define EXTEND_MSDOS		'M'
#define EXTEND_MACOS		'm'
#define EXTEND_OS9		'9'
#define EXTEND_OS2		'2'
#define EXTEND_OS68K		'K'
#define EXTEND_OS386		'3'		/* OS-9000??? */
#define EXTEND_HUMAN		'H'
#define EXTEND_CPM		'C'
#define EXTEND_FLEX		'F'
#define EXTEND_RUNSER		'R'
/*	this OS type is not official */
#define EXTEND_TOWNSOS		'T'
#define EXTEND_XOSK		'X'

enum
{
  ARCLZH_STORED = 0,

  ARCLZH_LZHED,		/* LZH base-tag */
  ARCLZH_LZHED_LH0,	/* -lh0- (ARCLZH_STORED) */
  ARCLZH_LZHED_LH1,	/* -lh1- */
  ARCLZH_LZHED_LH2,	/* -lh2- */
  ARCLZH_LZHED_LH3,	/* -lh3- */
  ARCLZH_LZHED_LH4,	/* -lh4- */
  ARCLZH_LZHED_LH5,	/* -lh5- */
  ARCLZH_LZHED_LZS,	/* -lzs- */
  ARCLZH_LZHED_LZ5,	/* -lz5- */
  ARCLZH_LZHED_LZ4,	/* -lz4- (ARCLZH_STORED) */
  ARCLZH_LZHED_LHD,	/* -lhd- (Directory, No compression data) */
  ARCLZH_LZHED_LH6,	/* -lh6- */
  ARCLZH_LZHED_LH7,	/* -lh7- */
};

static char *get_ptr;
#define setup_get(PTR)		(get_ptr = (PTR))
#define get_byte()		(*get_ptr++ & 0xff)
#define skip_byte()		get_ptr++
static char data[LZHEADER_STRAGE];

static unsigned short
get_word(void)
{
  int b0, b1;

  b0 = get_byte();
  b1 = get_byte();
  return (b1 << 8) + b0;
}

static long
get_longword(void)
{
  long b0, b1, b2, b3;

  b0 = get_byte();
  b1 = get_byte();
  b2 = get_byte();
  b3 = get_byte();
  return (b3 << 24) + (b2 << 16) + (b1 << 8) + b0;
}

static int
arclzhfs_name_check(const char *path)
{
  if ((path = strrchr(path, '.')) == NULL)
    return 0;
  path++;
  return strnicmp(path, "lzh", 3) == 0 && path[3] == '\0';
}

static void
fdskip(int fd, int len)
{
  lseek(fd, len, 1);
}

static void
pathfix(char *path, int len)
{
  while (--len > 0) {
    if (path[len] == '\\' || path[len] == '\377')
      path[len] = '/';
  }
}

static void
path_mac2unix(char *path, int len)
{
  while (--len > 0) {
    if (path[len] == ':')
      path[len] = '/';
  }
}

static int
fdgetc(int fd)
{
  char c;
  if (read(fd, &c, 1) < 0)
    return EOF;
  return c & 0xff;
}

static int
calc_sum(char *p, int len)
{
  int sum;

  for (sum = 0; len; len--)
    sum += *p++;

  return sum & 0xff;
}

static int
arclzhfs_install_arcfs(struct super_block *sb)
{
  struct arcfs_entry *entry;
  struct arcfs_super_block *asb =
    (struct arcfs_super_block *)sb->u.generic_sbp;
  int fd, ret;
  char dirname[FILENAME_LENGTH];
  char filename[FILENAME_LENGTH], *fname;
  int dir_length, name_length;
  int i;
  char *ptr;
  int header_level;
  char method_id[5];
  long compsize, origsize;
  int extend_type;
  int isdir;
  int header_size;
  int checksum;
  int unix_mode, unix_uid, unix_gid;
  time_t unix_time;
  unsigned long timestamp;

  if ((fd = open(asb->arcfile, 0, 0)) < 0) {
    printk(KERN_ERR
	   "%s: Can't open archive file for mounting. (errno=%d)\n",
	   asb->arcfile, errno);
    return -errno;
  }
  ret = 0;

  for (;;) {
    dir_length = 0;
    dirname[0] = '\0';
    name_length = 0;
    filename[0] = '\0';
    unix_mode = -1;
    unix_uid = unix_gid = 0;
    unix_time = (time_t)0;
    memset(data, 0, sizeof(data));

    if ((header_size = fdgetc(fd)) == EOF)
      break;
    if (header_size < I_NAME)
      break;

    if (read(fd, data + I_HEADER_CHECKSUM, header_size - 1) < header_size - 1)
      break;
    setup_get(data + I_HEADER_LEVEL);
    header_level = get_byte();

    if (header_level != 0 &&
	header_level != 1 &&
	header_level != 2) {
      ret = -EINVAL;
      goto out;
    }

    if (header_level != 2) {
      if (read(fd, data + header_size, 2) < 2)
	break;
    }

    setup_get(data + I_HEADER_CHECKSUM);
    checksum = get_byte();

    memcpy(method_id, data + I_METHOD, sizeof(method_id));

    setup_get(data + I_PACKED_SIZE);
    compsize = get_longword();
    origsize = get_longword();
    timestamp = get_longword(); /* last_modified_stamp */
    skip_byte(); /* attribute */

    if ((header_level = get_byte()) != 2) {
      if (calc_sum(data + I_METHOD, header_size) != checksum) {
	printk(KERN_ERR "LHA: Checksum error\n");
	ret = -EINVAL;
	break;
      }

      name_length = get_byte();
      for (i = 0; i < name_length; i ++)
	filename[i] =(char)get_byte();
      filename[name_length] = '\0';
    } else {
      unix_time = timestamp;
      filename[0] = '\0';
      name_length = 0;
    }

    if (header_size - name_length >= 24) {
      /* EXTEND FORMAT */
      get_word(); /* crc */
      extend_type = get_byte();
    } else if(header_size - name_length == 22) {
      /* Generic with CRC */
      get_word(); /* crc */
      extend_type = EXTEND_GENERIC;
    } else if(header_size - name_length == 20) {
      /* Generic no CRC */
      extend_type = EXTEND_GENERIC;
    } else {
      printk(KERN_ERR "%s: LZH file?\n", asb->arcfile);
      break;
    }

    if (extend_type == EXTEND_UNIX && header_level == 0) {
      skip_byte();			/* minor_version */
      unix_time = get_longword();	/* unix_last_modified_stamp */
      unix_mode = get_word();		/* unix_mode */
      unix_uid = get_word();		/* unix_uid */
      unix_gid = get_word();		/* unix_gid */
      goto parse_ok;
    }

    if (header_level > 0) {
      /* Extend Header */
      if (header_level != 2)
	setup_get(data + header_size);
      ptr = get_ptr;
      while ((header_size = get_word()) != 0) {
	int ii;
	if (header_level != 2) {
	  if (data + sizeof(data) - get_ptr < header_size)
	    break;
	  if (read(fd, get_ptr, header_size) < header_size)
	    break;
	}

	ii = get_byte();
	switch (ii) {
	case 0:
	  /*
	   * header crc
	   */
	  setup_get(get_ptr + header_size - 3);
	  break;
	case 1:
	  /*
	   * filename
	   */
	  name_length = header_size - 3;
	  if (name_length >= sizeof(filename) - 1)
	    goto out;
	  for (i = 0; i < name_length; i++)
	    filename[i] =(char)get_byte ();
	  filename[name_length] = '\0';
	  break;
	case 2:
	  /*
	   * directory
	   */
	  dir_length = header_size - 3;
	  if (dir_length >=  sizeof(dirname) - 1)
	    goto out;
	  for (i = 0; i < dir_length; i++)
	    dirname[i] = (char)get_byte ();
	  dirname[dir_length] = '\0';
	  pathfix(dirname, dir_length);
	  break;
	case 0x40:
	  /*
	   * MS-DOS attribute
	   */
	  if (extend_type == EXTEND_MSDOS ||
	      extend_type == EXTEND_HUMAN ||
	      extend_type == EXTEND_GENERIC)
	    get_word(); /* attribute */
	  break;
	case 0x50:
	  /*
	   * UNIX permission
	   */
	  if (extend_type == EXTEND_UNIX)
	    unix_mode = get_word(); /* unix_mode */
	  break;
	case 0x51:
	  /*
	   * UNIX gid and uid
	   */
	  if (extend_type == EXTEND_UNIX) {
	    unix_gid = get_word();	/* unix_gid */
	    unix_uid = get_word();	/* unix_uid */
	  }
	  break;
	case 0x52:
	  /*
	   * UNIX group name
	   */
	  setup_get(get_ptr + header_size - 3);
	  break;
	case 0x53:
	  /*
	   * UNIX user name
	   */
	  setup_get(get_ptr + header_size - 3);
	  break;
	case 0x54:
	  /*
	   * UNIX last modified time
	   */
	  if (extend_type == EXTEND_UNIX)
	    unix_time = get_longword(); /* unix_last_modified_stamp */
	  break;
	default:
	  /*
	   * other headers
	   */
	  setup_get(get_ptr + header_size - 3);
	  break;
	}
      }
      if (header_level != 2 && get_ptr - ptr != 2)
	compsize -= get_ptr - ptr - 2;
    }

    if (dir_length) {
      name_length += dir_length;
      if (name_length >= sizeof(filename) - 1)
	break;
      strcat(dirname, filename);
      strcpy(filename, dirname);
    }

    switch (extend_type) {
    case EXTEND_MSDOS:
      pathfix(filename, name_length);
      break;

    case EXTEND_HUMAN:
      if(header_level == 2)
	unix_time = timestamp;
      else
	unix_time = arc_dos2unix_time(timestamp & 0xffff,
				      (timestamp>>16) & 0xffff);
      break;

#ifdef OSK
    case EXTEND_OS68K:
    case EXTEND_XOSK:
#endif
    case EXTEND_UNIX:
      break;

    case EXTEND_MACOS:
      path_mac2unix(filename, name_length);
      unix_time = arc_dos2unix_time(timestamp & 0xffff,
				    (timestamp>>16) & 0xffff);
      break;

    default:
      pathfix(filename, name_length);
      if (header_level == 2)
	unix_time = timestamp;
      else
	unix_time = arc_dos2unix_time(timestamp & 0xffff,
				      (timestamp>>16) & 0xffff);
    }

  parse_ok:
    if (strncmp("-lhd-", method_id, 5) == 0)
      isdir = 1;
    else
      isdir = 0;

    for (i = 0; lzh_methods[i]; i++)
      if (strncmp(method_id, lzh_methods[i], sizeof(method_id)) == 0)
	break;
    if (!lzh_methods[i])
      break;

    if ((fname = arcfs_strip_topdir(filename, NULL)) == NULL) {
      fdskip(fd, compsize);
      continue;
    }

    PRINTK(("lzh %s\n", fname));
    if ((entry = arcfs_add_entry(sb, fname)) == NULL)
      break;

    if (unix_mode != -1)
      entry->mode = unix_mode;
    else if (isdir)
      entry->mode = 0555 | S_IFDIR;
    else
      entry->mode = 0444 | S_IFREG;
    entry->nlink = 1;
    entry->uid = unix_uid;
    entry->gid = unix_gid;
    entry->rdev = 0;
    entry->atime = entry->mtime = entry->ctime = unix_time;
    entry->size = origsize;
    entry->type = i + ARCLZH_LZHED + 1;
    entry->compsize = compsize;
    entry->offset = lseek(fd, 0, 1);
    fdskip(fd, compsize);
  }

 out:
  close(fd);
  if (ret >= 0) {
    MOD_INC_USE_COUNT;
  }
  return ret;
}

static void
arclzhfs_uninstall_arcfs(struct super_block *sb)
{
  PRINTK(("arclzhfs_uninstall()\n"));
  MOD_DEC_USE_COUNT;
}

static long
arc_read_func(char *buf, long size, void *user_val)
{
  struct arc_file *arc_filp = (struct arc_file *)user_val;
  if (arc_filp->pos + size > arc_filp->size)
    size = arc_filp->size - arc_filp->pos;
  if (size <= 0)
    return 0;
  return read(arc_filp->fd, buf, size);
}

static struct arc_file *
lzh_arc_open(struct arcfs_entry *entry)
{
  struct arc_file *arc_filp;
  struct arcfs_super_block *asb;
  int err;

  if (entry->type == ARCLZH_STORED)
    return default_arc_open(entry);

  asb = ENTRY_ARCFS(entry);
  if ((arc_filp = (struct arc_file *)kmalloc(sizeof(struct arc_file),
					     GFP_KERNEL)) == NULL) {
    err = ENOMEM;
    goto err1;
  }

  if ((arc_filp->fd = open(asb->arcfile, 0, 0)) < 0) {
    err = errno;
    goto err2;
  }

  if (lseek(arc_filp->fd, entry->offset, 0) < 0) {
    err = errno;
    goto err3;
  }

  arc_filp->offset = entry->offset;
  arc_filp->pos = 0;
  arc_filp->size = entry->size;
  arc_filp->type = entry->type;

  arc_filp->decoder =
    (void *)open_unlzh_handler(arc_read_func,
			      lzh_methods[arc_filp->type - ARCLZH_LZHED - 1],
			      entry->compsize,
			      arc_filp->size, arc_filp);

  if (arc_filp->decoder == NULL)
    err = ENOMEM;
  else
    return arc_filp;

 err3:
  close(arc_filp->fd);
 err2:
  kfree(arc_filp);
 err1:
  return ERR_PTR(-err);
}

static int
lzh_arc_rewind(struct arc_file *arc_filp)
{
  if (arc_filp->type == ARCLZH_STORED) {
    arc_filp->pos = 0;
    if (lseek(arc_filp->fd, arc_filp->offset, 0) < 0)
      return -errno;
    return 0;
  }
  if (arc_filp->decoder != NULL)
    close_unlzh_handler((UNLZHHandler)arc_filp->decoder);
  arc_filp->decoder = NULL;

  if (lseek(arc_filp->fd, arc_filp->offset, 0) < 0)
    return -errno;
  arc_filp->pos = 0;
  arc_filp->decoder =
    (void *)open_unlzh_handler(arc_read_func,
			       lzh_methods[arc_filp->type - ARCLZH_LZHED - 1],
			       arc_filp->entry->compsize,
			       arc_filp->size, arc_filp);
  if (arc_filp->decoder == NULL)
    return -ENOMEM;
  return 0;
}

static ssize_t
lzh_arc_read(struct arc_file *arc_filp, char *buff, size_t len)
{
  long n;

  if (arc_filp->type == ARCLZH_STORED)
    default_arc_read(arc_filp, buff, len);
  if (arc_filp->decoder == NULL)
    return -EINVAL;
  n = unlzh((UNLZHHandler)arc_filp->decoder, buff, len);
  if (n <= 0)
    return 0;
  arc_filp->pos += n;
  return n;
}

static loff_t
lzh_arc_seek(struct arc_file *arc_filp, loff_t pos)
{
  int err;
  ssize_t skip;

  if (arc_filp->pos == pos)
    return pos;

  if (arc_filp->pos > pos) {
    err = lzh_arc_rewind(arc_filp);
    if (err < 0)
      return err;
  }

  skip = arc_skip_read(arc_filp, lzh_arc_read, pos - arc_filp->pos);
  if (skip < 0)
    return skip;
  return arc_filp->pos;
}
	     
static int
lzh_arc_close(struct arc_file *arc_filp)
{
  int ret;

  if (arc_filp->decoder != NULL)
    close_unlzh_handler((UNLZHHandler)arc_filp->decoder);
  ret = close(arc_filp->fd);
  kfree(arc_filp);
  return ret;
}

static struct arcfs_module arclzhfs =
{
  arclzhfs_name_check,
  arclzhfs_install_arcfs,
  arclzhfs_uninstall_arcfs,

  lzh_arc_open,
  lzh_arc_seek,
  lzh_arc_read,
  lzh_arc_close,

  NULL,
};

int
init_module(void)
{
  PRINTK(("regist lzh fs\n"));
  return arcfs_register(&arclzhfs);
}

void
cleanup_module(void)
{
  PRINTK(("unregist lzh fs\n"));
  arcfs_unregister(&arclzhfs);
}
