#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/slab.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static int errno;

#include "arcfs.h"
#include "mblock.h"

static struct arcfs_module *arcfs_module_list = NULL;

#define PATH_SEP '/'
#define IS_PATH_SEP(x) ((x) == PATH_SEP)

/*
 * arcfs_expand_file_name() removes "/./" and "/../" from the path name.
 * It also strips beginning with "/".
 */
static char *
arcfs_expand_file_name(char* name)
{
  char *p, *o;

  /* "//" => "/" */
  p = o = name;
  while (*p) {
    if (IS_PATH_SEP(p[0]) && IS_PATH_SEP(p[1])) {
      p++;
      continue;
    }
    *o++ = *p++;
  }
  /* strip tail of the path separator */
  while (name < o - 1 && IS_PATH_SEP(o[-1]))
    o--;
  *o = '\0';

  /* Now canonicalize by removing /. and /foo/.. if they appear.  */
  o = p = name;
  while (*p) {
    if (!IS_PATH_SEP(*p))
      *o++ = *p++;
    else if (p[1] == '.' && (IS_PATH_SEP(p[2]) || p[2] == '\0'))
      p += 2;
    else if (p[1] == '.' && p[2] == '.'
	     && (IS_PATH_SEP(p[3]) || p[3] == '\0')) {
      while (o != name && (--o) && !IS_PATH_SEP(*o))
	;
      p += 3;
    }
    else
      *o++ = *p++;
  }
  *o = '\0';

  /* strip root */
  while (IS_PATH_SEP(*name))
    name++;

  return name;
}

void *
arcfs_alloc_memory(struct arcfs_super_block *asb, size_t len)
{
  return mblock_alloc((MBlock *)&asb->mem, len);
}

struct arcfs_entry *
arcfs_alloc_entry(struct super_block *sb, char *filename, int len)
{
  struct arcfs_entry *entry;
  struct arcfs_super_block *asb =
    (struct arcfs_super_block *)sb->u.generic_sbp;

  entry = (struct arcfs_entry *)
    arcfs_alloc_memory(asb,
		       sizeof(struct arcfs_entry) + len + 1);
  memset(entry, 0, sizeof(struct arcfs_entry) + len + 1);
  memcpy(entry->name, filename, len);
  entry->next = entry->sub = NULL;
  entry->sb = sb;
  entry->ino = ARCENT2INO(entry);
  entry->dev = sb->s_dev;
  return entry;
}

static struct arcfs_entry *
__arcfs_add_entry(struct super_block *sb,
		  struct arcfs_entry **base, char *path)
{
  char *filename;
  int flen;
  struct arcfs_entry *entry, *prev;

  /* split path into filename/path */
  filename = path;
  if ((path = strchr(filename, PATH_SEP)) == NULL)
    flen = strlen(filename);
  else {
    flen = path - filename;
    path++; /* skip PATH_SEP */
  }

  /* lookup filename */
  prev = NULL;
  for (entry = *base; entry; prev = entry, entry = entry->next) {
    if (strncmp(entry->name, filename, flen) == 0 &&
	(filename[flen] == PATH_SEP || filename[flen] == '\0'))
      break; /* found */
  }

  if (entry == NULL) {
    if ((entry = arcfs_alloc_entry(sb, filename, flen)) == NULL)
      return NULL;
    if (prev == NULL) {
      *base = entry;
    } else {
      while (prev->next)
	prev = prev->next;
      prev->next = entry;
    }
  }
  if (path == NULL)
    return entry;
  entry->mode |= S_IFDIR;
  return __arcfs_add_entry(sb, &entry->sub, path);
}

/*
 * arcfs_add_entry() adds the path into arcfs_entry.
 */
struct arcfs_entry *
arcfs_add_entry(struct super_block *sb, char *path)
{
  return __arcfs_add_entry(sb,
		&((struct arcfs_super_block *)sb->u.generic_sbp)->root->sub,
		arcfs_expand_file_name(path));
}

static struct arcfs_entry *
__arcfs_find_entry(struct arcfs_entry *base,  char *path)
{
  char *filename;
  int flen;
  struct arcfs_entry *entry;

  /* split path into filename/path */
  filename = path;
  if ((path = strchr(filename, PATH_SEP)) == NULL)
    flen = strlen(filename);
  else {
    flen = path - filename;
    path++; /* skip PATH_SEP */
  }

  /* lookup filename */
  for (entry = base; entry; entry = entry->next) {
    if (strncmp(entry->name, filename, flen) == 0 &&
	(entry->name[flen] == PATH_SEP || entry->name[flen] == '\0'))
      break;
  }

  if (path == NULL || entry == NULL)
    return entry;
  return __arcfs_find_entry(entry->sub, path);
}

/*
 * alloc_arcfs_entry() adds the path into arcfs_entry.
 */
struct arcfs_entry *
arcfs_find_entry(struct super_block *sb, char *path)
{
  return __arcfs_find_entry(
	((struct arcfs_super_block *)sb->u.generic_sbp)->root->sub,
	arcfs_expand_file_name(path));
}

int
arcfs_register(struct arcfs_module *module)
{
  struct arcfs_module ** tmp;

  if (!module)
    return -EINVAL;
  if (module->next)
    return -EBUSY;
  tmp = &arcfs_module_list;
  while (*tmp) {
    if (*tmp == module)
      return -EBUSY;
    tmp = &(*tmp)->next;
  }
  *tmp = module;
  return 0;
}

int
arcfs_unregister(struct arcfs_module *module)
{
  struct arcfs_module **tmp;

  tmp = &arcfs_module_list;
  while (*tmp) {
    if (module == *tmp) {
      *tmp = module->next;
      module->next = NULL;
      return 0;
    }
    tmp = &(*tmp)->next;
  }
  return -EINVAL;
}

struct arcfs_module *arcfs_find_module(const char *path)
{
  struct arcfs_module *p;

  for (p = arcfs_module_list; p; p = p->next)
    if (p->name_check(path))
      return p;
  return NULL;
}

struct arc_file *default_arc_open(struct arcfs_entry *entry)
{
  struct arcfs_super_block *asb;
  struct arc_file *arc_filp;
  int err;

  asb = ENTRY_ARCFS(entry);
  if ((arc_filp = (struct arc_file *)kmalloc(sizeof(struct arc_file),
					     GFP_KERNEL)) == NULL) {
    err = ENOMEM;
    goto err1;
  }
  if ((arc_filp->fd = open(asb->arcfile, 0, 0)) < 0) {
    err = errno;
    goto err2;
  }
  if (lseek(arc_filp->fd, entry->offset, 0) < 0) {
    err = errno;
    goto err3;
  }
  arc_filp->offset = entry->offset;
  arc_filp->pos = 0;
  arc_filp->size = entry->size;
  arc_filp->type = 0;
  arc_filp->decoder = NULL;
  return arc_filp;

 err3:
  close(arc_filp->fd);
 err2:
  kfree(arc_filp);
 err1:
  return ERR_PTR(-err);
}

loff_t
default_arc_seek(struct arc_file *arc_filp, loff_t pos)
{
  pos = lseek(arc_filp->fd, arc_filp->offset + pos, 0);
  if(pos < 0)
    return -errno;
  arc_filp->pos = pos - arc_filp->offset;
  return arc_filp->pos;
}

ssize_t
default_arc_read(struct arc_file *arc_filp, char *buff, size_t len)
{
  ssize_t n;
  if ((n = read(arc_filp->fd, buff, len)) < 0)
    return -errno;
  arc_filp->pos += n;
  return n;
}

int
default_arc_close(struct arc_file *arc_filp)
{
  int ret;
  ret = close(arc_filp->fd);
  kfree(arc_filp);
  return ret;
}

ssize_t
arc_skip_read(struct arc_file *arc_filp,
	      int (* arc_read)(struct arc_file *arc_filp,
			       char *buff, size_t len),
	      size_t len)
{
  char buff[PAGE_SIZE];
  int n;
  ssize_t total;

  total = 0;
  while (len > 0) {
    n = len;
    if (n > sizeof(buff))
      n = sizeof(buff);
    n = arc_read(arc_filp, buff, n);
    if (n < 0)
      return -errno;
    if (n == 0)
      break;
    len -= n;
    total += n;
  }
  return total;
}

/* Convert a MS-DOS time/date pair to a UNIX date (seconds since 1 1 70). */
time_t
arc_dos2unix_time(unsigned short time, unsigned short date)
{
  int month,year;
  time_t secs;
  extern struct timezone sys_tz;
  /* Linear day numbers of the respective 1sts in non-leap years. */
  static int day_n[] = { 0,31,59,90,120,151,181,212,243,273,304,334,0,0,0,0 };

  if (time == 0 && date == 0)
    return 0;

  month = ((date >> 5) & 15)-1;
  year = date >> 9;
  secs = (time & 31)*2+60*((time >> 5) & 63)+(time >> 11)*3600+86400*
    ((date & 31)-1+day_n[month]+(year/4)+year*365-((year & 3) == 0 &&
						   month < 2 ? 1 : 0)+3653);
  /* days since 1.1.70 plus 80's leap day */
  secs += sys_tz.tz_minuteswest*60;
  if (sys_tz.tz_dsttime)
    secs -= 3600;
  return secs;
}


/* strip "./" or "../" */
char *
arcfs_strip_topdir(char *fname, int *len_ret)
{
  int flen = strlen(fname);

  while (flen > 0) {
    if (fname[0] == '.' && fname[1] == '/') {
      fname += 2;
      flen -= 2;
      continue;
    }

    if (fname[0] == '.' && fname[1] == '.' && fname[2] == '/') {
      fname += 3;
      flen -= 3;
      continue;
    }
    break;
  }

  if (flen <= 0)
    return NULL;
  if (len_ret)
    *len_ret = flen;
  return fname;
}
