#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/ctype.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static int errno;

#include "arcfs.h"
#include "mblock.h"

#define TARBLKSIZ 512
#define TARHDRSIZ 512
#define  NAMSIZ      100
#define  TUNMLEN      32
#define  TGNMLEN      32

union record {
  char charptr[TARHDRSIZ];
  struct {
    char    name[NAMSIZ];
    char    mode[8];
    char    uid[8];
    char    gid[8];
    char    size[12];
    char    mtime[12];
    char    chksum[8];
    char    linkflag;
    char    linkname[NAMSIZ];
    char    magic[8];
    char    uname[TUNMLEN];
    char    gname[TGNMLEN];
    char    devmajor[8];
    char    devminor[8];
  } header;
};

static int
arctarfs_name_check(const char *path)
{
  if ((path = strrchr(path, '.')) == NULL)
    return 0;
  path++;
  return strnicmp(path, "tar", 3) == 0 && path[3] == '\0';
}

static long
octal_value(char *s, int len)
{
  long val;

  while (len > 0 && !isdigit((int)(unsigned char)*s)) {
    s++;
    len--;
  }

  val = 0;
  while (len > 0 && isdigit((int)(unsigned char)*s)) {
    val = ((val<<3) | (*s - '0'));
    s++;
    len--;
  }
  return val;
}

static int
tar_checksum(char *hdr)
{
  int i;

  long recorded_sum;
  long unsigned_sum;		/* the POSIX one :-) */
  long signed_sum;		/* the Sun one :-( */

  recorded_sum = octal_value(hdr + 148, 8);
  unsigned_sum = 0;
  signed_sum = 0;
  for (i = 0; i < TARBLKSIZ; i++) {
    unsigned_sum += 0xFF & hdr[i];
    signed_sum   += hdr[i];
  }

  /* Adjust checksum to count the "chksum" field as blanks.  */
  for (i = 0; i < 8; i++) {
    unsigned_sum -= 0xFF & hdr[148 + i];
    signed_sum   -= hdr[i];
  }
  unsigned_sum += ' ' * 8;
  signed_sum   += ' ' * 8;

  return unsigned_sum == recorded_sum || signed_sum == recorded_sum;
}

static int
arctarfs_install_arcfs(struct super_block *sb)
{
  union record hdr;
  long size, sizeb;
  struct arcfs_entry *entry;
  int fd;
  int isdir;
  struct arcfs_super_block *asb =
    (struct arcfs_super_block *)sb->u.generic_sbp;

  if ((fd = open(asb->arcfile, 0, 0)) < 0) {
    printk(KERN_ERR
	   "%s: Can't open archive file for mounting. (errno=%d)\n",
	   asb->arcfile, errno);
    return -errno;
  }

  MOD_INC_USE_COUNT;

  for (;;) {
    char *fname;
    int flen;

    if (read(fd, hdr.charptr, TARHDRSIZ) != TARHDRSIZ)
      break;
    if (hdr.header.name[0] == '\0')
      break;
    if (!tar_checksum(hdr.charptr))
      break;

    size = octal_value(hdr.header.size, 12);
    sizeb = (((size) + (TARBLKSIZ-1)) & ~(TARBLKSIZ-1));

    if ((fname = arcfs_strip_topdir(hdr.header.name, &flen)) == NULL) {
      lseek(fd, sizeb, 1);
      continue; /* empty file name */
    }

    isdir = (size == 0 && fname[flen - 1] == '/');
    PRINTK(("tar %s\n", fname));
    entry = arcfs_add_entry(sb, fname);

    if (entry == NULL)
      break;

    entry->mode = octal_value(hdr.header.mode, 8);
    entry->nlink = 1;
    entry->uid = octal_value(hdr.header.uid, 8);
    entry->gid = octal_value(hdr.header.gid, 8);
    entry->rdev = 0;

    if (S_ISLNK(entry->mode)) {
      entry->size = strlen(hdr.header.linkname);
      entry->symlink = (char *)arcfs_alloc_memory(asb, entry->size + 1);
      if (entry->symlink == NULL)
	break;
      memcpy(entry->symlink, hdr.header.linkname, entry->size);
      entry->symlink[entry->size] = '\0';
    } else
      entry->size = size;
    entry->atime = octal_value(hdr.header.mtime, 12);
    entry->mtime = octal_value(hdr.header.mtime, 12);
    entry->ctime = octal_value(hdr.header.mtime, 12);

    entry->type = 0;
    entry->compsize = entry->size;
    entry->offset = lseek(fd, 0, 1);

    lseek(fd, sizeb, 1);
  }
  close(fd);
  return 0;
}

static void
arctarfs_uninstall_arcfs(struct super_block *sb)
{
  MOD_DEC_USE_COUNT;
}

static struct arcfs_module arctarfs =
{
  arctarfs_name_check,
  arctarfs_install_arcfs,
  arctarfs_uninstall_arcfs,

  default_arc_open,
  default_arc_seek,
  default_arc_read,
  default_arc_close,

  NULL,
};

int
init_module(void)
{
  PRINTK(("regist tar fs\n"));
  return arcfs_register(&arctarfs);
}

void
cleanup_module(void)
{
  PRINTK(("unregist tar fs\n"));
  arcfs_unregister(&arctarfs);
}
