#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/slab.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static int errno;

#include "arcfs.h"
#include "mblock.h"
#include "zip.h"

#define TARBLKSIZ 512
#define TARHDRSIZ 512
#define  NAMSIZ      100
#define  TUNMLEN      32
#define  TGNMLEN      32

union record {
  char charptr[TARHDRSIZ];
  struct {
    char    name[NAMSIZ];
    char    mode[8];
    char    uid[8];
    char    gid[8];
    char    size[12];
    char    mtime[12];
    char    chksum[8];
    char    linkflag;
    char    linkname[NAMSIZ];
    char    magic[8];
    char    uname[TUNMLEN];
    char    gname[TGNMLEN];
    char    devmajor[8];
    char    devminor[8];
  } header;
};

static int
arctgzfs_name_check(const char *path)
{
  int len = strlen(path);

  if (len >= 5 && strnicmp(path + len - 4, ".tgz", 4) == 0)
    return 1;
  if (len >= 8 && strnicmp(path + len - 7, ".tar.gz", 7) == 0)
    return 1;

  return 0;
}

static long
octal_value(char *s, int len)
{
  long val;

  while (len > 0 && !isdigit((int)(unsigned char)*s)) {
    s++;
    len--;
  }

  val = 0;
  while (len > 0 && isdigit((int)(unsigned char)*s)) {
    val = ((val<<3) | (*s - '0'));
    s++;
    len--;
  }
  return val;
}

static int
tar_checksum(char *hdr)
{
  int i;

  long recorded_sum;
  long unsigned_sum;		/* the POSIX one :-) */
  long signed_sum;		/* the Sun one :-( */

  recorded_sum = octal_value(hdr + 148, 8);
  unsigned_sum = 0;
  signed_sum = 0;
  for (i = 0; i < TARBLKSIZ; i++) {
    unsigned_sum += 0xFF & hdr[i];
    signed_sum   += hdr[i];
  }

  /* Adjust checksum to count the "chksum" field as blanks.  */
  for (i = 0; i < 8; i++) {
    unsigned_sum -= 0xFF & hdr[148 + i];
    signed_sum   -= hdr[i];
  }
  unsigned_sum += ' ' * 8;
  signed_sum   += ' ' * 8;

  return unsigned_sum == recorded_sum || signed_sum == recorded_sum;
}

#define GZIP_ASCIIFLAG		(1u<<0)
#define GZIP_MULTIPARTFLAG	(1u<<1)
#define GZIP_EXTRAFLAG		(1u<<2)
#define GZIP_FILEFLAG		(1u<<3)
#define GZIP_COMMFLAG		(1u<<4)
#define GZIP_ENCFLAG		(1u<<5)
static int
arc_fdgetc(int fd) {
  char c;
  if (read(fd, &c, 1) <= 0)
    return -1;
  return c & 0xff;
}
static int
skip_gzip_header(int fd)
{
  unsigned char flags;
  int m1, method;

  /* magic */
  m1 = arc_fdgetc(fd);
  if (m1 != 0x1f)
    return -1;
  if (arc_fdgetc(fd) != 0x8b)
    return -1;

  /* method */
  method = arc_fdgetc(fd);

  /* flags */
  flags = arc_fdgetc(fd);
  if (flags & GZIP_ENCFLAG)
    return -1;
  /* time */
  arc_fdgetc(fd); arc_fdgetc(fd); arc_fdgetc(fd); arc_fdgetc(fd);

  arc_fdgetc(fd);		/* extra flags */
  arc_fdgetc(fd);		/* OS type */

  if (flags & GZIP_MULTIPARTFLAG) {
    /* part number */
    arc_fdgetc(fd); arc_fdgetc(fd);
  }

  if (flags & GZIP_EXTRAFLAG) {
    unsigned short len;
    int i;

    /* extra field */

    len = arc_fdgetc(fd);
    len |= ((unsigned short)arc_fdgetc(fd)) << 8;
    for (i = 0; i < len; i++)
      arc_fdgetc(fd);
  }

  if (flags & GZIP_FILEFLAG) {
    /* file name */
    int c;

    do {
      c = arc_fdgetc(fd);
      if(c == -1)
	return -1;
    } while(c != '\0');
  }

  if (flags & GZIP_COMMFLAG) {
    /* comment */
    int c;

    do {
      c = arc_fdgetc(fd);
      if (c == -1)
	return -1;
    } while(c != '\0');
  }

  return method;
}

static long
tgz_read_func(char *buf, long size, void *user_val)
{
  struct arc_file *arc_filp = (struct arc_file *)user_val;
  return read(arc_filp->fd, buf, size);
}

static off_t
tgz_skip_inflate(InflateHandler decoder, off_t size)
{
  char buff[1024];
  off_t n, total = 0;
  while (size > 0) {
    n = size;
    if (n > sizeof(buff))
       n = sizeof(buff);
    n = zip_inflate(decoder, buff, n);
    if (n <= 0)
      break;
    size -= n;
    total += n;
  }
  return total;
}

static int
arctgzfs_install_arcfs(struct super_block *sb)
{
  union record hdr;
  long size, sizeb;
  struct arcfs_entry *entry;
  int isdir;
  int flen;
  char *fname;
  struct arcfs_super_block *asb =
    (struct arcfs_super_block *)sb->u.generic_sbp;
  struct arc_file *arc_filp;
  int err;
  int method;
  off_t offset;

  if ((arc_filp = (struct arc_file *)kmalloc(sizeof(struct arc_file),
					     GFP_KERNEL)) == NULL) {
    err = ENOMEM;
    goto err1;
  }
  memset(arc_filp, 0, sizeof(struct arc_file));

  if ((arc_filp->fd = open(asb->arcfile, 0, 0)) < 0) {
    printk(KERN_ERR
	   "%s: Can't open archive file for mounting. (errno=%d)\n",
	   asb->arcfile, errno);
    return -errno;
  }

  method = skip_gzip_header(arc_filp->fd);
  if (method < 0) {
    printk(KERN_ERR
	   "%s: Not a valid gzip'ed file\n", asb->arcfile);
    err = EINVAL;
    goto err2;
  }
  if (method != 8) {
    printk(KERN_ERR
	   "%s: gzip method %d is not supported\n", asb->arcfile, method);
    err = EINVAL;
    goto err2;
  }

  if ((arc_filp->decoder = (void *)open_inflate_handler(tgz_read_func,
							arc_filp)) == NULL) {
    err = ENOMEM;
    goto err2;
  }

  offset = 0;
  for (;;) {
    if (zip_inflate((InflateHandler)arc_filp->decoder,
		    hdr.charptr, TARHDRSIZ) != TARHDRSIZ)
      break;
    if (hdr.header.name[0] == '\0')
      break;
    if (!tar_checksum(hdr.charptr))
      break;

    offset += TARHDRSIZ;
    size = octal_value(hdr.header.size, 12);
    sizeb = (((size) + (TARBLKSIZ-1)) & ~(TARBLKSIZ-1));

    if ((fname = arcfs_strip_topdir(hdr.header.name, &flen)) == NULL) {
      offset += tgz_skip_inflate((InflateHandler)arc_filp->decoder, sizeb);
      continue; /* empty file name */
    }

    isdir = (size == 0 && flen > 0 && fname[flen - 1] == '/');
    PRINTK(("tgz %s\n", fname));
    entry = arcfs_add_entry(sb, fname);

    if (entry == NULL)
      break;

    entry->mode = octal_value(hdr.header.mode, 8);
    entry->nlink = 1;
    entry->uid = octal_value(hdr.header.uid, 8);
    entry->gid = octal_value(hdr.header.gid, 8);
    entry->rdev = 0;

    if (S_ISLNK(entry->mode)) {
      entry->size = strlen(hdr.header.linkname);
      entry->symlink = (char *)arcfs_alloc_memory(asb,
						  entry->size + 1);
      if (entry->symlink == NULL)
	break;
      memcpy(entry->symlink, hdr.header.linkname, entry->size);
      entry->symlink[entry->size] = '\0';
    }
    else
      entry->size = size;
    entry->atime = octal_value(hdr.header.mtime, 12);
    entry->mtime = octal_value(hdr.header.mtime, 12);
    entry->ctime = octal_value(hdr.header.mtime, 12);

    entry->offset = offset;	/* start of body */
    offset += tgz_skip_inflate((InflateHandler)arc_filp->decoder, sizeb);
  }
  close_inflate_handler((InflateHandler)arc_filp->decoder);
  close(arc_filp->fd);
  kfree(arc_filp);
  MOD_INC_USE_COUNT;
  return 0;

 err2:
  close(arc_filp->fd);
  kfree(arc_filp);

 err1:
  return -err;
}

static struct arc_file *
tgz_arc_open(struct arcfs_entry *entry)
{
  struct arc_file *arc_filp;
  struct arcfs_super_block *asb;
  int err;

  asb = ENTRY_ARCFS(entry);
  if ((arc_filp = (struct arc_file *)kmalloc(sizeof(struct arc_file),
					     GFP_KERNEL)) == NULL) {
    err = ENOMEM;
    goto err1;
  }
  memset(arc_filp, 0, sizeof(arc_filp));

  if ((arc_filp->fd = open(asb->arcfile, 0, 0)) < 0) {
    err = errno;
    goto err2;
  }

  if (skip_gzip_header(arc_filp->fd) != 8) {
    err = EIO;
    goto err3;
  }

  arc_filp->decoder = (void *)open_inflate_handler(tgz_read_func,
						   arc_filp);
  if (arc_filp->decoder == NULL) {
    err = ENOMEM;
    goto err3;
  }

  tgz_skip_inflate((InflateHandler)arc_filp->decoder, entry->offset);
  arc_filp->offset = entry->offset;
  arc_filp->pos = 0;
  arc_filp->size = entry->size;
  return arc_filp;

 err3:
  close(arc_filp->fd);
 err2:
  kfree(arc_filp);
 err1:
  return ERR_PTR(-err);
}

static int
tgz_arc_rewind(struct arc_file *arc_filp)
{
  if (arc_filp->decoder == NULL)
    return -EIO;
  close_inflate_handler((InflateHandler)arc_filp->decoder);
  arc_filp->decoder = NULL;
  if (lseek(arc_filp->fd, 0, 0) < 0)
    return -errno;

  if (skip_gzip_header(arc_filp->fd) != 8) {
    return -EIO;
  }

  arc_filp->decoder = (void *)open_inflate_handler(tgz_read_func,
						   arc_filp);
  if (arc_filp->decoder == NULL) {
    return -EIO;
  }
  tgz_skip_inflate((InflateHandler)arc_filp->decoder, arc_filp->offset);
  arc_filp->pos = 0;
  return 0;
}

static ssize_t
tgz_arc_read(struct arc_file *arc_filp, char *buff, size_t len)
{
  long n;

  if (arc_filp->decoder == NULL)
    return -EINVAL;
  if (arc_filp->pos + len > arc_filp->size)
    len = arc_filp->size - arc_filp->pos;
  if (len <= 0)
    return 0;
  n = zip_inflate((InflateHandler)arc_filp->decoder, buff, len);
  if (n <= 0)
    return 0;
  arc_filp->pos += n;
  return n;
}

static loff_t
tgz_arc_seek(struct arc_file *arc_filp, loff_t pos)
{
  int err;
  ssize_t skip;

  if (arc_filp->pos == pos)
    return pos;

  if (arc_filp->pos > pos) {
    err = tgz_arc_rewind(arc_filp);
    if (err < 0)
      return err;
  }
  skip = tgz_skip_inflate((InflateHandler)arc_filp->decoder, pos - arc_filp->pos);
  arc_filp->pos += skip;
  return arc_filp->pos;
}

static int
tgz_arc_close(struct arc_file *arc_filp)
{
  int ret;
  if (arc_filp->decoder != NULL)
    close_inflate_handler((InflateHandler)arc_filp->decoder);
  arc_filp->decoder = NULL;
  ret = close(arc_filp->fd);
  kfree(arc_filp);
  return ret;
}

static void
arctgzfs_uninstall_arcfs(struct super_block *sb)
{
  MOD_DEC_USE_COUNT;
}

static struct arcfs_module arctgzfs =
{
  arctgzfs_name_check,
  arctgzfs_install_arcfs,
  arctgzfs_uninstall_arcfs,

  tgz_arc_open,
  tgz_arc_seek,
  tgz_arc_read,
  tgz_arc_close,

  NULL,
};

int
init_module(void)
{
  PRINTK(("regist tgz fs\n"));
  return arcfs_register(&arctgzfs);
}

void
cleanup_module(void)
{
  PRINTK(("unregist tgz fs\n"));
  arcfs_unregister(&arctgzfs);
}
