#include <linux/module.h>
#include <linux/fs.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/slab.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static int errno;

#include <linux/limits.h>

#include "arcfs.h"
#include "mblock.h"
#include "zip.h"
#include "explode.h"

#define LOCSIG		0x04034b50L
#define EXTLOCSIG	0x08074b50L
#define CENTSIG		0x02014b50L
#define ENDCENTSIG	0x06054b50L

/* extra-field ID values, all little-endian: */
#define MAX_EXTRA_LEN 256
#define EF_IZUNIX    0x5855    /* Info-ZIP's old Unix[1] ("UX") */
#define EF_IZUNIX2   0x7855    /* Info-ZIP's new Unix[2] ("Ux") */
#define EF_TIME      0x5455    /* universal timestamp ("UT") */

struct locsig_header
{
  char version[2];	/* Version needed to extract */
  char flags[2];	/* General purpose bit flag */
  char method[2];	/* Compression method */;
  char modtime[2];	/* Last mod file time */
  char moddate[2];	/* Last mod file date */
  char crc[4];		/* CRC-32 */
  char csize[4];	/* Compressed size */
  char usize[4];	/* Uncompressed size */
  char flen[2];		/* Filename length */
  char elen[2];		/* Extra field length */
  /* filename */
  /* extra field */
  /* data */
};

struct centsig_header
{
  char madeby[2];	/* Version made by */
  char version[2];	/* Version needed to extract */
  char flags[2];	/* General purpose bit flag */
  char method[2];	/* Compression method */
  char modtime[2];	/* Last mod file time */
  char moddate[2];	/* Last mod file date */
  char crc[4];		/* CRC-32 */
  char csize[4];	/* Compressed size */
  char usize[4];	/* Uncompressed size */
  char flen[2];		/* Filename length (f) */
  char elen[2];		/* Extra field length (e) */
  char clen[2];		/* File comment length (c) */
  char dnum[2];		/* Disk number start */
  char iattr[2];	/* Internal file attributes */
  char eattr[4];	/* External file attributes */
  char offset[4];	/* Relative offset of locnal header */
  /* Filename */
  /* Extra field */
  /* File comment */
};

enum
{
  ARCZIP_STORED = 0,
  ARCZIP_SHRUNKED,
  ARCZIP_REDUCED1,
  ARCZIP_REDUCED2,
  ARCZIP_REDUCED3,
  ARCZIP_REDUCED4,
  ARCZIP_IMPLODED_LIT8,
  ARCZIP_IMPLODED_LIT4,
  ARCZIP_IMPLODED_NOLIT8,
  ARCZIP_IMPLODED_NOLIT4,
  ARCZIP_DEFLATED
};

#define FS_FAT_           0    /* filesystem used by MS-DOS, OS/2, Win32 */
#define AMIGA_            1
#define VMS_              2
#define UNIX_             3
#define VM_CMS_           4
#define ATARI_            5    /* what if it's a minix filesystem? [cjh] */
#define FS_HPFS_          6    /* filesystem used by OS/2 (and NT 3.x) */
#define MAC_              7
#define Z_SYSTEM_         8
#define CPM_              9
#define TOPS20_           10
#define FS_NTFS_          11   /* filesystem used by Windows NT */
#define QDOS_             12
#define ACORN_            13   /* Archimedes Acorn RISC OS */
#define FS_VFAT_          14   /* filesystem used by Windows 95, NT */
#define MVS_              15
#define BEOS_             16   /* hybrid POSIX/database filesystem */
#define TANDEM_           17   /* Tandem/NSK */

static unsigned short get_short(char *s)
{
    unsigned char *p = (unsigned char *)s;
    return ((unsigned short)p[0] |
	    (unsigned short)p[1]<<8);
}

static unsigned long get_long(char *s)
{
    unsigned char *p = (unsigned char *)s;
    return ((unsigned long)p[0] |
	    (unsigned long)p[1]<<8 |
	    (unsigned long)p[2]<<16 |
	    (unsigned long)p[3]<<24);
}

static int arczipfs_name_check(const char *path)
{
  if((path = strrchr(path, '.')) == NULL)
    return 0;
  path++;
  return (strnicmp(path, "zip", 3) == 0 || strnicmp(path, "jar", 3) == 0)
    && path[3] == '\0';
}

static void fdskip(int fd, int len)
{
  lseek(fd, len, 1);
}

static void path_dos2unix(char *path)
{
  while(*path)
    {
      if(*path == '\\')
	*path = '/';
      path++;
    }
}

static int
arc_method_type(int flags, int method)
{
  /* Choose compress method */
  switch (method) {
  case 0: /* The file is stored (no compression) */
    method = ARCZIP_STORED;
    break;
  case 1: /* The file is Shrunk */
    method = ARCZIP_SHRUNKED;
    break;
  case 2: /* The file is Reduced with compression factor 1 */
    method = ARCZIP_REDUCED1;
    break;
  case 3: /* The file is Reduced with compression factor 2 */
    method = ARCZIP_REDUCED2;
    break;
  case 4: /* The file is Reduced with compression factor 3 */
    method = ARCZIP_REDUCED3;
    break;
  case 5: /* The file is Reduced with compression factor 4 */
    method = ARCZIP_REDUCED4;
    break;
  case 6: /* The file is Imploded */
    if (flags & 4) {
      if (flags & 2)
	method = ARCZIP_IMPLODED_LIT8;
      else
	method = ARCZIP_IMPLODED_LIT4;
    } else if (flags & 2)
      method = ARCZIP_IMPLODED_NOLIT8;
    else
      method = ARCZIP_IMPLODED_NOLIT4;
    break;
  case 7: /* Reserved for Tokenizing compression algorithm */
    method = -1;
    break;
  case 8: /* The file is Deflated */
    method = ARCZIP_DEFLATED;
    break;
  default:
    method = -1;
    break;
  }

  switch (method) {
  case ARCZIP_SHRUNKED:
  case ARCZIP_REDUCED1:
  case ARCZIP_REDUCED2:
  case ARCZIP_REDUCED3:
  case ARCZIP_REDUCED4:
    method = -1; /* Unsupported */
    break;
  }

  return method;
}

static int mapattr(unsigned long attr, int hostnum)
{
    switch(hostnum)
      {
      case UNIX_:
      case VMS_:
      case ACORN_:
      case ATARI_:
      case BEOS_:
      case QDOS_:
	return attr >> 16;

      case AMIGA_:
	attr = (unsigned)(attr>>17 & 7);   /* Amiga RWE bits */
	return (unsigned)(attr<<6 | attr<<3 | attr);

        /* all remaining cases:  expand MSDOS read-only bit into write perms */
      case FS_FAT_:
      case FS_HPFS_:
      case FS_NTFS_:
      case MAC_:
      case TOPS20_:
      default:
	/* read-only bit --> write perms; subdir bit --> dir exec bit */
	attr = !(attr & 1) << 1  |  (attr & 0x10) >> 4;
	return (0444 | attr<<6 | attr<<3 | attr);
      }
    return 0;  /* dummy */
}

static void parse_extra_field(char *extra, int extra_len,
			      struct arcfs_entry *entry)
{
  int flen, fid;
  while(extra_len > 4)
    {
      fid = get_short(extra);
      flen = get_short(extra + 2);
      if(extra_len - 4 < flen)
	break;
      extra += 4;
      extra_len -= 4;

      switch(fid) {
      case EF_IZUNIX2:
	if(flen >= 4)
	  {
	    entry->uid = get_short(extra + 0);
	    entry->gid = get_short(extra + 2);
	  }
	break;

      case EF_IZUNIX:
	if(flen >= 8)
	  {
	    entry->atime = get_long(extra + 0);
	    entry->mtime = get_long(extra + 4);
	    if(flen >= 12)
	      {
		entry->uid = get_short(extra + 8);
		entry->gid = get_short(extra + 10);
	      }
	  }
	break;

      case EF_TIME:
	if(flen >= 1)
	  {
	    unsigned int eb_idx = 1;
	    unsigned int flags;
#define EB_UT_FL_MTIME    (1 << 0)      /* mtime present */
#define EB_UT_FL_ATIME    (1 << 1)      /* atime present */
#define EB_UT_FL_CTIME    (1 << 2)      /* ctime present */

	    flags = extra[0] & 0xff;
	    if(flags & EB_UT_FL_MTIME)
	      {
		if(eb_idx + 4 <= flen)
		  {
		    entry->mtime = get_long(extra + eb_idx);
		    eb_idx += 4;
		  }
	      }
	    if(flags & EB_UT_FL_ATIME)
	      {
		if(eb_idx + 4 <= flen)
		  {
		    entry->atime = get_long(extra + eb_idx);
		    eb_idx += 4;
		  }
	      }
	    if(flags & EB_UT_FL_CTIME)
	      {
		if(eb_idx + 4 <= flen)
		  {
		    entry->ctime = get_long(extra + eb_idx);
		  }
	      }
	  }
	break;
      }
      extra += flen;
      extra_len -= flen;
    }
}

static int
arczipfs_install_arcfs(struct super_block *sb)
{
  struct arcfs_entry *entry;
  int fd, ret;
  struct arcfs_super_block *asb =
    (struct arcfs_super_block *)sb->u.generic_sbp;
  char buff[PATH_MAX], *fname;
  struct locsig_header lhdr;
  struct centsig_header chdr;
  unsigned short flags;
  unsigned short flen, elen, clen;
  int method;
  umode_t dirmode;
  int magic;
  int n;
  char extra[MAX_EXTRA_LEN];

  if ((fd = open(asb->arcfile, 0, 0)) < 0) {
    printk(KERN_ERR
	   "%s: Can't open archive file for mounting. (errno=%d)\n",
	   asb->arcfile, errno);
    return -errno;
  }

  ret = 0;
  for (;;) {
    if (read(fd, buff, 4) != 4) {
      ret = -EINVAL;
      break;
    }
    magic = get_long(buff);
    switch (magic) {
    case LOCSIG:
      if (read(fd, (char *)&lhdr, sizeof(lhdr)) != sizeof(lhdr)) {
	ret = -EINVAL;
	goto out;
      }

      flags = get_short(lhdr.flags);
      flen  = get_short(lhdr.flen);
      elen  = get_short(lhdr.elen);
      method = get_short(lhdr.method);
      method = arc_method_type(flags, method);

      /* read filename */
      n = flen;
      if (n >= sizeof(buff) - 1)
	n = sizeof(buff) - 1;
      if (read(fd, buff, n) != n) {
	ret = -EINVAL;
	break;
      }
      buff[n] = '\0';
      fdskip(fd, flen - n);

      /* extra field */
      if (elen <= sizeof(extra))
	read(fd, extra, elen);
      else
	fdskip(fd, elen);

      if ((fname = arcfs_strip_topdir(buff, &n)) == NULL) {
	fdskip(fd, get_long(lhdr.csize));
	break;
      }

      path_dos2unix(fname);
      if (n > 0 && fname[n - 1] == '/')
	dirmode = 0555 | S_IFDIR;
      else
	dirmode = 0444 | S_IFREG;

      PRINTK(("zip %s\n", fname));
      if ((entry = arcfs_add_entry(sb, fname)) == NULL) {
	fdskip(fd, get_long(lhdr.csize));
	break;
      }
      entry->mode = 0444 | dirmode;
      entry->nlink = 1;
      entry->uid = 0;
      entry->gid = 0;
      entry->rdev = 0;
      entry->atime = entry->mtime = entry->ctime =
	arc_dos2unix_time(get_short(lhdr.modtime),
			  get_short(lhdr.moddate));

      entry->type = method;
      entry->compsize = get_long(lhdr.csize);
      if (method == -1)
	entry->size = entry->compsize;
      else
	entry->size = get_long(lhdr.usize);

      if (elen <= sizeof(extra))
	parse_extra_field(extra, elen, entry);

      entry->offset = lseek(fd, 0, 1);
      fdskip(fd, entry->compsize);
      break;

    case EXTLOCSIG: /* ignored */
      if (read(fd, buff, 20) != 20) {
	ret = -EINVAL;
	goto out;
      }
      break; /* End of EXTLOCSIG */

    case CENTSIG:
      if (read(fd, (char *)&chdr, sizeof(chdr)) != sizeof(chdr)) {
	ret = -EINVAL;
	goto out;
      }

      flen  = get_short(chdr.flen);
      elen  = get_short(chdr.elen);
      clen  = get_short(chdr.clen);

      /* read filename */
      n = flen;
      if (n >= sizeof(buff) - 1)
	n = sizeof(buff) - 1;
      if (read(fd, buff, n) != n) {
	ret = -EINVAL;
	break;
      }
      buff[n] = '\0';
      fdskip(fd, flen - n);

      /* extra field */
      if (elen <= sizeof(extra))
	read(fd, extra, elen);
      else
	fdskip(fd, elen);

      /* skip comment field */
      fdskip(fd, clen);

      if ((entry = arcfs_find_entry(sb, buff)) == NULL)
	break; /* Skip CENTSIG */
      entry->mode = mapattr(get_long(chdr.eattr), chdr.madeby[1]);
      if (elen <= sizeof(extra))
	parse_extra_field(extra, elen, entry);
      break; /* End of CENTSIG */

    case ENDCENTSIG:
      goto out; /* Done */
    default:
      ret = -EINVAL;
      goto out;
    }
  }
 out:
  close(fd);
  if (ret >= 0) {
    MOD_INC_USE_COUNT;
  }
  return ret;
}

static void
arczipfs_uninstall_arcfs(struct super_block *sb)
{
  MOD_DEC_USE_COUNT;
}

static long
arc_read_func(char *buf, long size, void *user_val)
{
  struct arc_file *arc_filp = (struct arc_file *)user_val;
  if (arc_filp->pos + size > arc_filp->size)
    size = arc_filp->size - arc_filp->pos;
  if (size <= 0)
    return 0;
  return read(arc_filp->fd, buf, size);
}

static struct arc_file *zip_arc_open(struct arcfs_entry *entry)
{
  struct arc_file *arc_filp;
  struct arcfs_super_block *asb;
  int err;

  if (entry->type == ARCZIP_STORED)
    return default_arc_open(entry);

  asb = ENTRY_ARCFS(entry);
  if ((arc_filp = (struct arc_file *)kmalloc(sizeof(struct arc_file),
					     GFP_KERNEL)) == NULL) {
    err = ENOMEM;
    goto err1;
  }

  if ((arc_filp->fd = open(asb->arcfile, 0, 0)) < 0) {
    err = errno;
    goto err2;
  }

  if (lseek(arc_filp->fd, entry->offset, 0) < 0) {
    err = errno;
    goto err3;
  }

  arc_filp->offset = entry->offset;
  arc_filp->pos = 0;
  arc_filp->size = entry->size;
  arc_filp->type = entry->type;

  switch (entry->type) {
  case ARCZIP_IMPLODED_LIT8:
  case ARCZIP_IMPLODED_LIT4:
  case ARCZIP_IMPLODED_NOLIT8:
  case ARCZIP_IMPLODED_NOLIT4:
    arc_filp->decoder =
      (void *)open_explode_handler(arc_read_func,
				   entry->type - ARCZIP_IMPLODED_LIT8,
				   entry->compsize, entry->size, arc_filp);
    break;

  case ARCZIP_DEFLATED:
    arc_filp->decoder = (void *)open_inflate_handler(arc_read_func, arc_filp);
    break;

  default:
    err = EINVAL;
    goto err3;
  }

  if (arc_filp->decoder == NULL)
    err = ENOMEM;
  else
    return arc_filp;

 err3:
  close(arc_filp->fd);
 err2:
  kfree(arc_filp);
 err1:
  return ERR_PTR(-err);
}

static int
zip_arc_rewind(struct arc_file *arc_filp)
{
  switch (arc_filp->type) {
  case ARCZIP_IMPLODED_LIT8:
  case ARCZIP_IMPLODED_LIT4:
  case ARCZIP_IMPLODED_NOLIT8:
  case ARCZIP_IMPLODED_NOLIT4:
    close_explode_handler((ExplodeHandler)arc_filp->decoder);
    break;
  case ARCZIP_DEFLATED:
    close_inflate_handler((InflateHandler)arc_filp->decoder);
    break;
  }
  arc_filp->decoder = NULL;

  if (lseek(arc_filp->fd, arc_filp->offset, 0) < 0)
    return -errno;
  arc_filp->pos = 0;

  switch (arc_filp->type) {
  case ARCZIP_IMPLODED_LIT8:
  case ARCZIP_IMPLODED_LIT4:
  case ARCZIP_IMPLODED_NOLIT8:
  case ARCZIP_IMPLODED_NOLIT4:
    arc_filp->decoder =
      (void *)open_explode_handler(arc_read_func,
				   arc_filp->type - ARCZIP_IMPLODED_LIT8,
				   arc_filp->entry->compsize,
				   arc_filp->size, arc_filp);
    if (arc_filp->decoder == NULL)
      return -ENOMEM;

  case ARCZIP_DEFLATED:
    arc_filp->decoder = (void *)open_inflate_handler(arc_read_func, arc_filp);
    if (arc_filp->decoder == NULL)
      return -ENOMEM;
    break;
  }
  return 0;
}

static ssize_t
zip_arc_read(struct arc_file *arc_filp, char *buff, size_t len)
{
  long n;

  switch (arc_filp->type) {
  case ARCZIP_STORED:
    return default_arc_read(arc_filp, buff, len);

  case ARCZIP_IMPLODED_LIT8:
  case ARCZIP_IMPLODED_LIT4:
  case ARCZIP_IMPLODED_NOLIT8:
  case ARCZIP_IMPLODED_NOLIT4:
    if (arc_filp->decoder == NULL)
      return -EIO;
    n = explode((ExplodeHandler)arc_filp->decoder, buff, len);
    break;
  case ARCZIP_DEFLATED:
    if (arc_filp->decoder == NULL)
      return -EIO;
    n = zip_inflate((InflateHandler)arc_filp->decoder, buff, len);
    break;
  default:
    return -EINVAL;
  }
  if (n <= 0)
    return 0;
  arc_filp->pos += n;
  return n;
}

static loff_t
zip_arc_seek(struct arc_file *arc_filp, loff_t pos)
{
  int err;
  ssize_t skip;

  if (arc_filp->pos == pos)
    return pos;

  if (arc_filp->pos > pos) {
    err = zip_arc_rewind(arc_filp);
    if (err < 0)
      return err;
  }

  skip = arc_skip_read(arc_filp, zip_arc_read, pos - arc_filp->pos);
  if (skip < 0)
    return skip;
  return arc_filp->pos;
}
	     
static int
zip_arc_close(struct arc_file *arc_filp)
{
  int ret;

  if (arc_filp->decoder != NULL) {
    switch (arc_filp->type) {
    case ARCZIP_IMPLODED_LIT8:
    case ARCZIP_IMPLODED_LIT4:
    case ARCZIP_IMPLODED_NOLIT8:
    case ARCZIP_IMPLODED_NOLIT4:
      close_explode_handler((ExplodeHandler)arc_filp->decoder);
      break;
    case ARCZIP_DEFLATED:
      close_inflate_handler((InflateHandler)arc_filp->decoder);
      break;
    }
  }
  ret = close(arc_filp->fd);
  kfree(arc_filp);
  return ret;
}

static struct arcfs_module arczipfs =
{
  arczipfs_name_check,
  arczipfs_install_arcfs,
  arczipfs_uninstall_arcfs,

  zip_arc_open,
  zip_arc_seek,
  zip_arc_read,
  zip_arc_close,

  NULL,
};

int
init_module(void)
{
  PRINTK(("regist zip fs\n"));
  return arcfs_register(&arczipfs);
}

void
cleanup_module(void)
{
  PRINTK(("unregist zip fs\n"));
  arcfs_unregister(&arczipfs);
}
